/*
 * Copyright (C) 2022  Aravinth Manivannan <realaravinth@batsense.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
use std::cell::RefCell;

use actix_identity::Identity;
use actix_web::http::header::ContentType;
use serde::{Deserialize, Serialize};
use tera::Context;
use uuid::Uuid;

use super::get_auth_middleware;

use crate::db::Site;
use crate::pages::dash::TemplateSiteEvent;
use crate::pages::errors::*;
use crate::settings::Settings;
use crate::AppCtx;

pub use super::*;

pub const DASH_GITEA_WEBHOOK_VIEW: TemplateFile =
    TemplateFile::new("dash_gitea_webhook_view", "pages/dash/gitea/view.html");

const SHOW_GITEA_WEBHOOK_SECRET_KEY: &str = "show_gitea_webhook_secret";

pub struct View {
    ctx: RefCell<Context>,
}

impl CtxError for View {
    fn with_error(&self, e: &ReadableError) -> String {
        self.ctx.borrow_mut().insert(ERROR_KEY, e);
        self.render()
    }
}

impl View {
    pub fn new(settings: &Settings, payload: Option<TemplateGiteaWebhook>) -> Self {
        let ctx = RefCell::new(context(settings));
        if let Some(payload) = payload {
            ctx.borrow_mut().insert(PAYLOAD_KEY, &payload);
        }

        Self { ctx }
    }

    pub fn show_gitea_webhook_secret(&mut self) {
        self.ctx
            .borrow_mut()
            .insert(SHOW_GITEA_WEBHOOK_SECRET_KEY, &true);
    }

    pub fn render(&self) -> String {
        TEMPLATES
            .render(DASH_GITEA_WEBHOOK_VIEW.name, &self.ctx.borrow())
            .unwrap()
    }
}

#[derive(Serialize, Deserialize, Clone, Debug, Eq, PartialEq)]
pub struct ViewOptions {
    show_gitea_webhook_secret: Option<bool>,
}

#[actix_web_codegen_const_routes::get(
    path = "PAGES.dash.gitea_webhook.view",
    wrap = "get_auth_middleware()"
)]
#[tracing::instrument(name = "Dashboard Gitea webhook webpage", skip(ctx, id))]
pub async fn get_view_site(
    ctx: AppCtx,
    id: Identity,
    path: web::Path<String>,
    query: web::Query<ViewOptions>,
) -> PageResult<impl Responder, View> {
    let auth_token = path.into_inner();
    let owner = id.identity().unwrap();

    let hook = ctx
        .db
        .get_webhook_with_owner(&auth_token, &owner)
        .await
        .map_err(|e| PageError::new(View::new(&ctx.settings, None), e))?;

    let payload = TemplateGiteaWebhook::new(&ctx, hook);
    let mut page = View::new(&ctx.settings, Some(payload));
    if let Some(true) = query.show_gitea_webhook_secret {
        page.show_gitea_webhook_secret();
    }
    let add = page.render();
    let html = ContentType::html();
    Ok(HttpResponse::Ok().content_type(html).body(add))
}

pub fn services(cfg: &mut web::ServiceConfig) {
    cfg.service(get_view_site);
}
